// $Id$

/*************************************************************************
 * XDAQ Components for Distributed Data Acquisition                      *
 * Copyright (C) 2000-2022, CERN.			                 *
 * All rights reserved.                                                  *
 * Authors: L. Orsini, D. Simelevicius					 *
 *                                                                       *
 * For the licensing terms see LICENSE.		                         *
 * For the list of contributors see CREDITS.   			         *
 *************************************************************************/

#include "CGICCTest.h"

#include <new>
#include <string>
#include <vector>
#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include "cgicc/CgiDefs.h"
#include "cgicc/Cgicc.h"
#include "cgicc/HTTPHTMLHeader.h"
#include "cgicc/HTMLClasses.h"

#if HAVE_SYS_UTSNAME_H
#  include <sys/utsname.h>
#endif

#if HAVE_SYS_TIME_H
#  include <sys/time.h>
#endif


// Function prototypes
void dumpEnvironment(const cgicc::CgiEnvironment& env,xgi::Output * out);
void dumpList(const cgicc::Cgicc& formData, xgi::Output * out);
void showForm(const cgicc::Cgicc& formData, xgi::Output * out);
void printForm(const cgicc::Cgicc& cgi, xgi::Output * out);

//
// provides factory method for instantion of SimpleWeb application
//
XDAQ_INSTANTIATOR_IMPL(CGICCTest)

CGICCTest::CGICCTest(xdaq::ApplicationStub * s)
	: xdaq::Application(s)
{	
	xgi::bind(this,&CGICCTest::Default, "Default");
}
	
void CGICCTest::Default(xgi::Input * in, xgi::Output * out ) 
{
        cgicc::Cgicc cgi(in);

	// This method can be invoked using Linux 'wget' command
	// e.g http://lxcmd101:1972/urn:xdaq-application:lid=23/setParameter?value=24
	//std::string method = toolbox::toString("/%s/Deafult",getApplicationDescriptor()->getURN().c_str());
	
  try {
#if HAVE_GETTIMEOFDAY
    timeval start;
    gettimeofday(&start, NULL);
#endif

    // Create a new Cgicc object containing all the CGI data
    //Cgicc cgi;
    
    // If the user wants to throw an exception, go ahead and do it
    if(cgi.queryCheckbox("throw") && ! cgi.queryCheckbox("restore"))
      throw std::runtime_error("User-requested Exception thrown in main()");
    
    // Output the HTTP headers for an HTML document, and the HTML 4.0 DTD info
    //*out << cgicc::HTTPHTMLHeader() << cgicc::HTMLDoctype(cgicc::HTMLDoctype::eStrict) << std::endl;
    out->getHTTPResponseHeader().addHeader("Content-Type", "text/html");
    *out << cgicc::HTMLDoctype(cgicc::HTMLDoctype::eStrict) << std::endl;
    *out << cgicc::html().set("lang", "en").set("dir", "ltr") << std::endl;

    // Set up the page's header and title.
    // I will put in lfs to ease reading of the produced HTML. 
    *out << cgicc::head() << std::endl;

    // Output the style sheet portion of the header
    *out << cgicc::style() << cgicc::comment() << std::endl;
    *out << styles;
    *out << cgicc::comment() << cgicc::style() << std::endl;

    *out << cgicc::title() << "GNU cgicc v" << cgi.getVersion() << " Test" 
	 << cgicc::title() << std::endl;

    *out << cgicc::head() << std::endl;
    
    // Start the HTML body
    *out << cgicc::body() << std::endl;

    *out << cgicc::h1() << "GNU cgi" << cgicc::span("cc").set("class","red")
	 << " v"<< cgi.getVersion() << " Test" << cgicc::h1() << std::endl;
    
    // Get a pointer to the environment
    const cgicc::CgiEnvironment& env = cgi.getEnvironment();
    
    // Generic thank you message
    *out << cgicc::comment() << "This page generated by cgicc for "
	 << env.getRemoteHost() << cgicc::comment() << std::endl;
    *out << cgicc::h4() << "Thanks for using cgi" << cgicc::span("cc").set("class", "red") 
	 << ", " << env.getRemoteHost() 
	 << '(' << env.getRemoteAddr() << ")!" << cgicc::h4() << std::endl;  
    
    // If the user wants to save the submission, do it
    if(cgi.queryCheckbox("save")) {
      // Make sure the save file is readable and writable by the CGI process
      cgi.save("save");
      *out << cgicc::p(cgicc::h2("Data Saved")) << std::endl;
      
      *out << cgicc::div().set("class", "notice") << std::endl;
      *out << "Your data has been saved, and may be restored (by anyone) "
	   << "via the same form." << std::endl << cgicc::div() << cgicc::p() << std::endl;
    }

    // If the user wants to restore from the last submission, do it
    if(cgi.queryCheckbox("restore")) {
      cgi.restore("save");
      *out << cgicc::p(cgicc::h2("Data Restored")) << std::endl;
      
      *out << cgicc::div().set("class", "notice") << std::endl;
      *out << "The data displayed has been restored from a file on disk."
	   << std::endl << cgicc::div() << cgicc::p() << std::endl;
    }
    
    // If the user requested a dump of the environment,
    // create a simple table showing the values of the 
    // environment variables
    if(cgi.queryCheckbox("showEnv"))
      dumpEnvironment(env,out);
    
    // If the user requested, print out the raw form data from 
    // the vector of FormEntries.  This will contain every 
    // element in the list.
    // This is one of two ways to get at form data, the other
    // being the use of Cgicc's getElement() methods.  
    if(cgi.queryCheckbox("showFE"))
      dumpList(cgi,out);
    
    // If the user requested data via Cgicc's getElement() methods, do it.
    // This is different than the use of the list of FormEntries 
    // because it requires prior knowledge of the name of form elements.
    // Usually they will be known, but you never know.
    if(cgi.queryCheckbox("showForm"))
      showForm(cgi,out);

    // Print out the form to do it again
    *out << cgicc::br() << std::endl;
    printForm(cgi,out);
    *out << cgicc::hr().set("class", "half") << std::endl;
    
    // Information on cgicc
    *out << cgicc::div().set("align","center").set("class","smaller") << std::endl;
    *out << "GNU cgi" << cgicc::span("cc").set("class","red") << " v";
    *out << cgi.getVersion() << cgicc::br() << std::endl;
    *out << "Compiled at " << cgi.getCompileTime();
    *out << " on " << cgi.getCompileDate() << cgicc::br() << std::endl;

    *out << "Configured for " << cgi.getHost();  
#if HAVE_UNAME
    struct utsname info;
    if(uname(&info) != -1) {
      *out << ". Running on " << info.sysname;
      *out << ' ' << info.release << " (";
      *out << info.nodename << ")." << std::endl;
    }
#else
    *out << "." << std::endl;
#endif

#if HAVE_GETTIMEOFDAY
    // Information on this query
    timeval end;
    gettimeofday(&end, NULL);
    long us = ((end.tv_sec - start.tv_sec) * 1000000)
      + (end.tv_usec - start.tv_usec);

    *out << cgicc::br() << "Total time for request = " << us << " us";
    *out << " (" << static_cast<double>(us/1000000.0) << " s)";
#endif

    // End of document
    *out << cgicc::div() << std::endl;
    *out << cgicc::body() << cgicc::html() << std::endl;

  }

  // Did any errors occur?
  catch(const std::exception& e) {

    // This is a dummy exception handler, as it doesn't really do
    // anything except print out information.

    // Reset all the HTML elements that might have been used to 
    // their initial state so we get valid output
	  cgicc::html::reset(); 	cgicc::head::reset(); 		cgicc::body::reset();
	  cgicc::title::reset(); 	cgicc::h1::reset(); 		cgicc::h4::reset();
	  cgicc::comment::reset(); 	cgicc::td::reset(); 		cgicc::tr::reset(); 
	  cgicc::table::reset();	cgicc::div::reset(); 	cgicc::p::reset(); 
	  cgicc::a::reset();		cgicc::h2::reset(); 		cgicc::colgroup::reset();

    // Output the HTTP headers for an HTML document, and the HTML 4.0 DTD info
    *out << cgicc::HTTPHTMLHeader() << cgicc::HTMLDoctype(cgicc::HTMLDoctype::eStrict) << std::endl;
    *out << cgicc::html().set("lang","en").set("dir","ltr") << std::endl;

    // Set up the page's header and title.
    // I will put in lfs to ease reading of the produced HTML. 
    *out << cgicc::head() << std::endl;

    // Output the style sheet portion of the header
    *out << cgicc::style() << cgicc::comment() << std::endl;
    *out << "body { color: black; background-color: white; }" << std::endl;
    *out << "hr.half { width: 60%; align: center; }" << std::endl;
    *out << "span.red, STRONG.red { color: red; }" << std::endl;
    *out << "div.notice { border: solid thin; padding: 1em; margin: 1em 0; "
	 << "background: #ddd; }" << std::endl;

    *out << cgicc::comment() << cgicc::style() << std::endl;

    *out << cgicc::title("GNU cgicc exception") << std::endl;
    *out << cgicc::head() << std::endl;
    
    *out << cgicc::body() << std::endl;
    
    *out << cgicc::h1() << "GNU cgi" << cgicc::span("cc", cgicc::set("class","red"))
	 << " caught an exception" << cgicc::h1() << std::endl; 
  
    *out << cgicc::div().set("align","center").set("class","notice") << std::endl;

    *out << cgicc::h2(e.what()) << std::endl;

    // End of document
    *out << cgicc::div() << std::endl;
    *out << cgicc::hr().set("class","half") << std::endl;
    *out << cgicc::body() << cgicc::html() << std::endl;
    
  }
	
	
}

// Print the form for this CGI
void
printForm(const cgicc::Cgicc& cgi, xgi::Output * out)
{
	// This method can be invoked using Linux 'wget' command
        // e.g http://lxcmd101:1972/urn:xdaq-application:lid=23/setParameter?value=24
        //std::string method = toolbox::toString("/%s/Deafult",getApplicationDescriptor()->getURN().c_str());

  *out << "<form method=\"post\" action=\"/" 
       << cgi.getEnvironment().getScriptName() << "/Default\">" << std::endl;
    
  *out << "<table>" << std::endl;

  *out << "<tr><td class=\"title\">Your name</td>"
       << "<td class=\"form\">"
       << "<input type=\"text\" name=\"name\" value=\"Uncle Bob\" />"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your salary in millions<br />(80-120)</td>"
       << "<td class=\"form\">"
       << "<input type=\"text\" name=\"bucks\" value=\"93\" />"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Hours you've wasted on the web</td>"
       << "<td class=\"form\">"
       << "<input type=\"text\" name=\"time\" value=\"100\" />"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your thoughts (on anything)</td>"
       << "<td class=\"form\">"
       << "<textarea name=\"thoughts\" rows=\"4\" cols=\"40\">"
       << "I don't have any!</textarea>" << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Are you hungry?</td>"
       << "<td class=\"form\">"
       << "<input type=\"checkbox\" name=\"hungry\" checked=\"checked\" />"
       << "Yes</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your favorite flavors of ice cream</td>"
       << "<td class=\"form\">"
       << "<select name=\"flavors\" multiple=\"multiple\">"
       << "<option value=\"cookie dough\">Cookie Dough</option>"
       << "<option value=\"rocky road\">Rocky Road</option>"
       << "<option value=\"chocolate\">Chocolate</option>"
       << "<option value=\"strawberry\">Strawberry</option>"
       << "<option value=\"vanilla\">Vanilla</option>"
       << "</select>" << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your hair color</td>"
       << "<td class=\"form\">"
       << "<select name=\"hair\">"
       << "<option value=\"blond\">Blond</option>"
       << "<option value=\"brown\">Brown</option>"
       << "<option value=\"red\">Red</option>"
       << "<option value=\"black\">Black</option>"
       << "<option value=\"white\">White</option>"
       << "<option value=\"green\">Green</option>"
       << "<option value=\"multicolored\">Multicolored</option>"
       << "</select>" << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your web browser</td>"
       << "<td class=\"form\">"
       << "<input type=\"radio\" name=\"browser\" value=\"Konqueror\""
       << " checked=\"checked\" />Konqeuror"
       << "<input type=\"radio\" name=\"browser\" value=\"Lynx\" />Lynx"
       << "<input type=\"radio\" name=\"browser\" value=\"Mozilla\" />Mozilla"
       << "<input type=\"radio\" name=\"browser\" value=\"IE\" />IE"
       << "<input type=\"radio\" name=\"browser\" value=\"Other\" />Other"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Your favorite authors</td>"
       << "<td class=\"form\">"
       << "<input type=\"checkbox\" name=\"authors\" value=\"O'Brian\" />"
       << "O'Brian"
       << "<input type=\"checkbox\" name=\"authors\" value=\"Feynman\" />" 
       << "Feynman"
       << "<input type=\"checkbox\" name=\"authors\" value=\"Camus\" />Camus"
       << "<input type=\"checkbox\" name=\"authors\" value=\"Conrad\" />Conrad"
       << "<input type=\"checkbox\" name=\"authors\" value=\"Vergil\" />Vergil"
       << "<input type=\"checkbox\" name=\"authors\" value=\"Plato\" />Plato"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">In the output, show</td>"
       << "<td class=\"form\">"
       << "<input type=\"checkbox\" name=\"showEnv\" checked=\"checked\" />"
       << "Data from CgiEnvironment<br />"
       << "<input type=\"checkbox\" name=\"showFE\" checked=\"checked\" />"
       << "All FormEntries<br />"
       << "<input type=\"checkbox\" name=\"showForm\" checked=\"checked\" />"
       << "Data from Cgicc"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Exception Handling</td>"
       << "<td class=\"form\">"
       << "<input type=\"checkbox\" name=\"throw\" />"
       << "Throw an exception to test error handling"
       << "</td></tr>" << std::endl;

  *out << "<tr><td class=\"title\">Save and Restore</td>"
       << "<td class=\"form\">"
       << "<input type=\"checkbox\" name=\"save\" />"
       <<" Save submission to a file<br />"
       << "<input type=\"checkbox\" name=\"restore\" />"
       << "Restore data from the last saved submission"
       << "</td></tr>" << std::endl;

  *out << "</table>" << std::endl;

  *out << "<div class=\"center\"><p>"
       << "<input type=\"submit\" name=\"submit\"  value=\"Submit\" />"
       << "<input type=\"reset\" value=\"Nevermind\" />"
       << "</p></div></form>" << std::endl;
}


// Print out a table of the CgiEnvironment
void
dumpEnvironment(const cgicc::CgiEnvironment& env, xgi::Output * out) 
{
  // This is just a brain-dead dump of information.
  // Almost all of this code is for HTML formatting
  *out << cgicc::h2("Environment information from CgiEnvironment") << std::endl;
  
  *out << cgicc::div().set("align","center") << std::endl;
  
  *out << cgicc::table() << std::endl;
  
  *out << cgicc::tr() << cgicc::td("Request Method").set("class","title") 
       << cgicc::td(env.getRequestMethod()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Path Info").set("class","title") 
       << cgicc::td(env.getPathInfo()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Path Translated").set("class","title") 
       << cgicc::td(env.getPathTranslated()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Script Name").set("class","title") 
       << cgicc::td(env.getScriptName()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("HTTP Referrer").set("class","title") 
       << cgicc::td(env.getReferrer()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("HTTP Cookie").set("class","title") 
       << cgicc::td(env.getCookies()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Query String").set("class","title") 
       << cgicc::td(env.getQueryString()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Content Length").set("class","title") 
       << cgicc::td().set("class","data") << env.getContentLength() 
       << cgicc::td() << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Post Data").set("class","title")
       << cgicc::td().set("class","data")
       << cgicc::pre(env.getPostData()).set("class","data") << cgicc::td() 
       << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Remote Host").set("class","title") 
       << cgicc::td(env.getRemoteHost()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Remote Address").set("class","title") 
       << cgicc::td(env.getRemoteAddr()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Authorization Type").set("class","title") 
       << cgicc::td(env.getAuthType()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Remote User").set("class","title") 
       << cgicc::td(env.getRemoteUser()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Remote Identification").set("class","title") 
       << cgicc::td(env.getRemoteIdent()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Content Type").set("class","title") 
       << cgicc::td(env.getContentType()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("HTTP Accept").set("class","title") 
       << cgicc::td(env.getAccept()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("User Agent").set("class","title") 
       << cgicc::td(env.getUserAgent()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Server Software").set("class","title") 
       << cgicc::td(env.getServerSoftware()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Server Name").set("class","title") 
       << cgicc::td(env.getServerName()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Gateway Interface").set("class","title") 
       << cgicc::td(env.getGatewayInterface()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Server Protocol").set("class","title") 
       << cgicc::td(env.getServerProtocol()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Server Port").set("class","title") 
       << cgicc::td().set("class","data") << env.getServerPort() 
       << cgicc::td() << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("HTTPS").set("class","title")
       << cgicc::td().set("class","data") << (env.usingHTTPS() ? "true" : "false")
       << cgicc::td() << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Redirect Request").set("class","title") 
       << cgicc::td(env.getRedirectRequest()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Redirect URL").set("class","title") 
       << cgicc::td(env.getRedirectURL()).set("class","data") << cgicc::tr() << std::endl;
  *out << cgicc::tr() << cgicc::td("Redirect Status").set("class","title") 
       << cgicc::td(env.getRedirectStatus()).set("class","data") << cgicc::tr() << std::endl;
  
  *out << cgicc::table() << cgicc::div() << std::endl;
}

// Print out the value of every form element
void
dumpList(const cgicc::Cgicc& formData, xgi::Output * out) 
{
  *out << cgicc::h2("Form Data via vector") << std::endl;
  
  *out << cgicc::div().set("align","center") << std::endl;
  
  *out << cgicc::table()<< std::endl;
  
  *out << cgicc::tr().set("class","title") << cgicc::td("Element Name") 
       << cgicc::td("Element Value") << cgicc::tr() << std::endl;
  
  // Iterate through the vector, and print out each value
  cgicc::const_form_iterator iter;
  for(iter = formData.getElements().begin(); 
      iter != formData.getElements().end(); 
      ++iter) {
    *out << cgicc::tr().set("class","data") << cgicc::td(iter->getName()) 
	 << cgicc::td(iter->getValue()) << cgicc::tr() << std::endl;
  }
  *out << cgicc::table() << cgicc::div() << std::endl;
}

// Print out information customized for each element
void
showForm(const cgicc::Cgicc& formData, xgi::Output * out) 
{

  // I am using an if statement to check if each element is found
  *out << cgicc::h2("Form Data via Cgicc") << std::endl;
  
  *out << cgicc::div().set("class","notice") << std::endl;

  //getElement
  cgicc::const_form_iterator name = formData.getElement("name");
  if(name != (*formData).end() && ! name->isEmpty())
    *out << "Your name is " << **name << '.' << cgicc::br() << std::endl;
  else
    *out << "You don't have a name." << cgicc::br() << std::endl;

  // getElement and getDoubleValue
  cgicc::const_form_iterator salary = formData.getElement("bucks");
  if(salary != (*formData).end() && ! salary->isEmpty())
    *out << "You make " << (*salary).getDoubleValue(80, 120) 
	 << " million dollars." << cgicc::br() << std::endl;
  else
    *out << "You don't have a salary." << cgicc::br() << std::endl;

  // getElement and getIntegerValue
  cgicc::const_form_iterator hours = formData.getElement("time");
  if(hours != (*formData).end() && ! (*hours).isEmpty())
    *out << "You've wasted " << (*hours).getIntegerValue() 
	 << " hours on the web." << cgicc::br() << std::endl;
  else
    *out << "You haven't wasted any time on the web." << cgicc::br() << std::endl;

  // getElement and getStrippedValue
  cgicc::const_form_iterator thoughts = formData.getElement("thoughts");
  if(thoughts != (*formData).end() && ! (*thoughts).isEmpty()) {
    std::string temp = (*thoughts).getStrippedValue();
    *out << "Your thoughts: " << temp << cgicc::br() << std::endl;
  }
  else
    *out << "You don't have any thoughts!?" << cgicc::br() << std::endl;
  
  // queryCheckbox
  if(formData.queryCheckbox("hungry"))
    *out << "You are hungry." << cgicc::br() << std::endl;
  else
    *out << "You are not hungry." << cgicc::br() << std::endl;

  // getElement
  std::vector<cgicc::FormEntry> flavors;
  formData.getElement("flavors", flavors);
  if(! flavors.empty()) {
    *out << "You like ";
    for(std::string::size_type i = 0; i < flavors.size(); i++) {
      *out << flavors[i].getValue();
      if(i < flavors.size() - 2)
	*out << ", ";
      else if(i == flavors.size() - 2)
	*out << " and ";
    }
    *out << " ice cream." << cgicc::br() << std::endl;
  }
  else
    *out << "You don't like ice cream!?" << cgicc::br() << std::endl;
  
  // getElement
  cgicc::const_form_iterator hair = formData.getElement("hair");
  if(hair != (*formData).end())
    *out << "Your hair is " << **hair << '.' << cgicc::br() << std::endl;
  else
    *out << "You don't have any hair." << cgicc::br() << std::endl;
  
  *out << "You surf the web with " << formData("browser") << '.' 
       << cgicc::br() << std::endl;
  
  // getElement
  std::vector<cgicc::FormEntry> authors;
  formData.getElement("authors", authors);
  if(! authors.empty()) {
    *out << "You like to read books by ";
    for(std::string::size_type i = 0; i < authors.size(); ++i) {
      *out << authors[i].getValue();
      if(i < authors.size() - 2)
	*out << ", ";
      else if(i == authors.size() - 2)
	*out << " and ";
    }
    *out << "." << cgicc::br() << std::endl;
  }
  else
    *out << "You don't like to read!?" << cgicc::br() << std::endl;
  
  *out << cgicc::div() << std::endl;
}
