#!/bin/bash
########################################################################
# build_xdaq.sh
#
# Build xdaq and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_10_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
	cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
	   ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

	Help. Supply <product-dir> for specific, detailed information on
	invoking ${prog}.

  <ssi-msg-config-options>

	Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

	The top-level UPS directory in which the source code distribution
	was unpacked. When specified with -h|--help, this enables us to
	provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

	Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
	build_script_usage
	exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
	[[ "$arg" ==  -* ]] && continue;
	product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
	# Bad usage.
	ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
	exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
	# Something wrong with <product-dir>.
	ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
	  "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
	"$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
	source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
	ensure_ups "${product_dir}" && \
	setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
	ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/bin"
  ${trace} verify_in_filesystem "${@}" "${pkgdir}/bin/xdaq.exe"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
	# Customize mbd_opts. *C*
	local mbd_opts=()
	${trace} is_built "$@" && \
	  ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=xdaq
origpkgver=16_20_0_1
pkgver=v${origpkgver}b # Append letters for local patches.
ssibuildshims_version=v2_10_00
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=core-release_${origpkgver}
pkgtarfile=${pkgtarname}.tar
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e26,e28,c7,c14,c15} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

	Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

	Command tracing and increased verbosity for source unpacking and patching
	operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   ${trace} setup_cmake --latest={--ge,3.15} || exit
#      (version requirement necessary for cmake --install).
# AND/OR
#   ${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${blddir}" \
  || ssi_die "unable to cd to build directory ${blddir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" "${tardir}/${pkgtarfile}" || exit

${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}/${pkgtarname}"
    
${trace} patch_source "${ssi_args[@]}" -p 0 -P "${patchdir}" ${patchdir}/*.patch || exit


# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#

# Set configure / compile flags, environment variables, etc.
export ExtraCFlags=""
export ExtraCCFlags="-std=c++14 -Wno-deprecated-declarations"

case "$basequal" in
e1[79]) CC='gcc'  CXX='g++';ExtraCCFlags="$ExtraCCFlags -std=c++17 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1";     type g++;;
e2[06]) CC='gcc'  CXX='g++';ExtraCCFlags="$ExtraCCFlags -std=c++17 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1";     type g++;;
e28) CC='gcc'  CXX='g++';ExtraCCFlags="$ExtraCCFlags -std=c++20 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1";     type g++;;
c[27])  CC='clang' CXX='clang++';ExtraCCFlags="$ExtraCCFlags -std=c++17 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1 -Wno-ignored-optimization-argument -Wno-unused-private-field -Wno-register"; type clang++;;
c1[45])  CC='clang' CXX='clang++';ExtraCCFlags="$ExtraCCFlags -std=c++17 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1 -Wno-ignored-optimization-argument -Wno-unused-private-field -Wno-register"; type clang++;;
*)       CXX='g++';     type g++;;
esac
case "$extraqual" in
prof)  ExtraCCFlags="$ExtraCCFlags -O3";;
debug) ExtraCCFlags="$ExtraCCFlags -O0";;
*)     ExtraCCFlags="$ExtraCCFlags -O3";;
esac

export CC
export CXX

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
#
# [[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
#   export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}
#
# For ctest:
#
# [[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
#   export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.


##################
# Configure. *C*

# e.g.
#   ${trace} ${sourcedir}/${pkgtarname}/configure \
#     --prefix="${pkgdir}" ... || ssi_die "configure failed"
# OR
#   ${trace} cmake <opts> ${DEBUG_CONFIG:+--verbose} \
#     "${sourcedir}/${pkgtarname}" || \
#     ssi_die "configure failed"
##################

externPackages="\
extern/nlohmannjson extern/xerces extern/asyncresolv extern/i2o extern/cgicc extern/log4cplus \
extern/xalan extern/mimetic extern/jansson"

Packages="\
  config xcept log/udpappender log/xmlappender toolbox xoap \
  xoap/filter xdata pt xgi i2o xdaq i2o/utils pt/http pt/fifo \
  executive hyperdaq \
  xrelay b2in/nub b2in/utils b2in/eventing \
"

rm -rf ../install
mkdir ../install
export XDAQ_ROOT=$PWD  # xdaq build/Makefiles need this -- especially since the setup
								  # of the fake product (above) sets to to the wrong location
export BUILD_HOME=$PWD
export PROJECT_NAME=core
export INSTALL_PREFIX=$PWD/../install
export BUILD_SUPPORT=build


##################
# Build. *C*

# e.g.
#   ${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
#     ssi_die "build failed"
##################

echo '#####################################################'
echo 'make Packages="$externPackages"' CXX=\"$CXX\"
echo '#####################################################'
for extpackage in $externPackages; do
	make -j$ncores Packages="$extpackage" CXX="$CXX" install || ssi_die "make failed. (\"externPackages\")"
	${trace} verify_in_filesystem -d "${@}" "${INSTALL_PREFIX}/x86_64_slc6/include/${extpackage}"
done
wait

echo '#####################################################'
echo 'make Packages="$Packages"' CXX=\"$CXX\"
echo '#####################################################'
make  Packages="$Packages"        CXX="$CXX" LD="$CXX" || ssi_die "make failed. (\"Packages\")"
wait

##################
# Install. *C*

# e.g.
#   ${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

daqdir=$PWD
for pack in $externPackages;do
  cd $daqdir
  cp -a $pack ${sourcedir}
done

for pack in $Packages;do
  cd $pack
  make install
  cd $daqdir
  cp -a $pack ${sourcedir}
done

cd $INSTALL_PREFIX/x86_64_*
mv doc share
cp -a $daqdir/build ./config
cp config/favicon.ico htdocs
find scripts -type f -exec mv {} bin \;
rm -rf scripts
mkdir etc
cp -p $daqdir/xdaq/etc/*.profile* etc
#echo Installing docs...

echo fixup libs??? cd lib - do ln -s

#
mv share ${product_dir}/${package}/${pkgver}
mv include ${product_dir}/${package}/${pkgver}
mkdir -p ${pkgdir}
mv bin ${pkgdir}
mv config ${pkgdir}
mv etc ${pkgdir}
mv htdocs ${pkgdir}
mv lib ${pkgdir}

echo Done with install.

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

echo Cleaning up...
cd "${pkgdir}"   # e.g. /home/ron/work/otsdaqPrj/myFW_workdir/xdaq/v14_4_0/Linux64bit+3.10-2.17-e14-prof

echo Wrapper scripts and config files...
find bin etc -type f | while read file; do
  [[ "$(file "${file}")" == *"text"* ]] && \
   perl -wapi\~ -e 's&\Q'"${pkgdir}"'\E&\${XDAQ_FQ_DIR}&g' "${file}"
done

#rm -rvf lib/pkgconfig lib/*.la
find . -name '*~' -exec rm -vf \{\} \;

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
